import time
from functools import wraps
from threading import Timer, Lock
from uuid import uuid4

from ovos_config.config import Configuration
from ovos_utils.log import LOG
from ovos_utils.messagebus import Message, FakeBus
from ovos_utils.network_utils import is_connected

from ovos_backend_client.api import DeviceApi, BackendType
from ovos_backend_client.exceptions import BackendDown, InternetDown, HTTPError
from ovos_backend_client.identity import IdentityManager
from ovos_backend_client.backends.selene import SELENE_API_URL


def is_backend_disabled():
    config = Configuration()
    if not config.get("server"):
        # missing server block implies disabling backend
        return True
    return config["server"].get("disabled") or False


def requires_backend(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        if not is_backend_disabled():
            return f(*args, **kwargs)
        return {}

    return decorated


def has_been_paired():
    """ Determine if this device has ever been paired with a backend

    Returns:
        bool: True if ever paired with backend (not factory reset)
    """
    # This forces a load from the identity file in case the pairing state
    # has recently changed
    ident = IdentityManager.load()
    # NOTE: even offline backend creates a dummy identity file
    return ident.uuid is not None and ident.uuid != ""


def is_paired(ignore_errors=True, url=None, version="v1", identity_file=None, backend_type=None):
    """Determine if this device is actively paired with a web backend

    Determines if the installation of Mycroft has been paired by the user
    with the backend system, and if that pairing is still active.

    Returns:
        bool: True if paired with backend
    """
    if is_backend_disabled():
        return True

    # check if pairing is valid
    api = DeviceApi(url=url, version=version, identity_file=identity_file, backend_type=backend_type)
    return api.identity.uuid and check_remote_pairing(ignore_errors, url=url, version=version,
                                                      identity_file=identity_file, backend_type=backend_type)


def check_remote_pairing(ignore_errors, url=None, version="v1", identity_file=None, backend_type=None):
    """Check that a basic backend endpoint accepts our pairing.

    Args:
        ignore_errors (bool): True if errors should be ignored when

    Returns:
        True if pairing checks out, otherwise False.
    """
    try:
        return bool(DeviceApi(url=url, version=version,
                              identity_file=identity_file, backend_type=backend_type).get())
    except HTTPError as e:
        if e.response.status_code == 401:
            return False
        error = e
    except Exception as e:
        error = e

    LOG.warning(f'Could not get device info: {error}')

    if ignore_errors:
        return False

    if isinstance(error, HTTPError):
        if is_connected():
            raise BackendDown from error
        else:
            raise InternetDown from error
    else:
        raise error


class PairingManager:
    poll_frequency = 5  # secs between checking server for activation

    def __init__(self, bus=None,
                 enclosure=None,
                 code_callback=None,
                 error_callback=None,
                 success_callback=None,
                 start_callback=None,
                 restart_callback=None,
                 end_callback=None,
                 pairing_url="home.mycroft.ai",
                 api_url=SELENE_API_URL,
                 version="v1",
                 identity_file=None,
                 backend_type=None):
        if enclosure:
            LOG.warning("enclosure argument has been deprecated, it is no longer used")
        self.pairing_url = pairing_url
        self.api_url = api_url
        self.restart_callback = restart_callback
        self.code_callback = code_callback
        self.error_callback = error_callback
        self.success_callback = success_callback
        self.start_callback = start_callback
        self.end_callback = end_callback

        self.bus = bus or FakeBus()
        self.api = DeviceApi(url=api_url, version=version,
                             identity_file=identity_file, backend_type=backend_type)
        self.data = None
        self.time_code_expires = None
        self.uuid = str(uuid4())
        self.activator = None
        self.activator_lock = Lock()
        self.activator_cancelled = False
        self.counter_lock = Lock()
        self.count = -1  # for repeating pairing code. -1 = not running
        self.num_failed_codes = 0

    def set_api_url(self, url,  version="v1", identity_file=None, backend_type=BackendType.SELENE):
        if not url.startswith("http"):
            url = f"http://{url}"
        self.api_url = url
        self.api = DeviceApi(url, version=version,
                             identity_file=identity_file,
                             backend_type=backend_type)

    def shutdown(self):
        with self.activator_lock:
            self.activator_cancelled = True
            if self.activator:
                self.activator.cancel()
        if self.activator:
            self.activator.join()

    def kickoff_pairing(self):
        self.data = None

        # Kick off pairing...
        with self.counter_lock:
            if self.count > -1:
                # We snuck in to this handler somehow while the pairing
                # process is still being setup.  Ignore it.
                LOG.debug("Ignoring call to kickoff_pairing")
                return
            # Not paired or already pairing, so start the process.
            self.count = 0

        LOG.debug("Kicking off pairing sequence")

        try:
            # Obtain a pairing code from the backend
            self.data = self.api.get_code(self.uuid)
            self.handle_pairing_code()

            # Keep track of when the code was obtained.  The codes expire
            # after 20 hours.
            self.time_code_expires = time.monotonic() + 72000  # 20 hours
        except Exception:
            time.sleep(10)
            # Call restart pairing here
            # Bail out after Five minutes (5 * 6 attempts at 10 seconds
            # interval)
            if self.num_failed_codes < 5 * 6:
                self.num_failed_codes += 1
                self.abort_and_restart(quiet=True)
            else:
                self.end_pairing('connection.error')
                self.num_failed_codes = 0
            return

        self.num_failed_codes = 0  # Reset counter on success

        if self.start_callback:
            self.start_callback()

        if not self.activator:
            self.__create_activator()

    def check_for_activate(self):
        """Method is called every 10 seconds by Timer. Checks if user has
        activated the device yet on home.mycroft.ai and if not repeats
        the pairing code every 60 seconds.
        """
        try:
            # Attempt to activate.  If the user has completed pairing on the,
            # backend, this will succeed.  Otherwise it throws and HTTPError()

            token = self.data.get("token")

            LOG.info(f"Attempting device activation @ {self.api_url}")
            login = self.api.activate(self.uuid, token)  # HTTPError() thrown
            if not login:
                raise ValueError("Received empty identity data!")
            LOG.info(f"Identity data received!: {login.get('uuid')}")

            # When we get here, the pairing code has been entered on the
            # backend and pairing can now be saved.
            # The following is kinda ugly, but it is really critical that we
            # get this saved successfully or we need to let the user know that
            # they have to perform pairing all over again at the website.
            try:
                IdentityManager.save(login)
            except Exception as e:
                LOG.debug("First save attempt failed: " + repr(e))
                time.sleep(2)
                try:
                    IdentityManager.save(login)
                except Exception as e2:
                    # Something must be seriously wrong
                    LOG.debug("Second save attempt failed: " + repr(e2))
                    self.abort_and_restart()
                    return

            # Assume speaking is the pairing code.  Stop TTS of that.
            self.bus.emit(Message("mycroft.audio.speech.stop"))

            self.bus.emit(Message("mycroft.paired", login))

            if self.success_callback:
                self.success_callback()

            # Un-mute.  Would have been muted during onboarding for a new
            # unit, and not dangerous to do if pairing was started
            # independently.
            self.bus.emit(Message("mycroft.mic.unmute", None))

        except HTTPError:
            # speak pairing code every 60th second
            with self.counter_lock:
                if self.count == 0:
                    self.handle_pairing_code()
                self.count = (self.count + 1) % 6

            if time.monotonic() > self.time_code_expires:
                # After 20 hours the token times out.  Restart
                # the pairing process.
                with self.counter_lock:
                    self.count = -1
                self.data = None
                if self.restart_callback:
                    self.restart_callback()
            else:
                # trigger another check in 10 seconds
                self.__create_activator()
        except Exception as e:
            LOG.debug("Unexpected error: " + repr(e))
            self.abort_and_restart()

    def end_pairing(self, error_dialog):
        """Resets the pairing and don't restart it.

        Arguments:
            error_dialog: Reason for the ending of the pairing process.
        """
        self.bus.emit(Message("mycroft.mic.unmute", None))
        self.data = None
        self.count = -1
        if self.end_callback:
            self.end_callback(error_dialog)

    def abort_and_restart(self, quiet=False):
        # restart pairing sequence
        LOG.debug("Aborting Pairing")
        # Reset state variables for a new pairing session
        with self.counter_lock:
            self.count = -1
        self.activator = None
        self.data = None  # Clear pairing code info
        LOG.info("Restarting pairing process")
        if self.error_callback:
            self.error_callback(quiet)
        self.bus.emit(Message("mycroft.not.paired",
                              data={'quiet': quiet}))

    def __create_activator(self):
        # Create a timer that will poll the backend in 10 seconds to see
        # if the user has completed the device registration process
        with self.activator_lock:
            if not self.activator_cancelled:
                self.activator = Timer(self.poll_frequency,
                                       self.check_for_activate)
                self.activator.daemon = True
                self.activator.start()

    def handle_pairing_code(self):
        """Log pairing code."""
        code = self.data.get("code")
        LOG.info("Pairing code: " + code)
        # emit info message, allows PHAL plugins to perform actions
        # eg. the mk1 faceplate scrolls the code
        self.bus.emit(Message("mycroft.pairing.code", self.data))
        if self.code_callback:
            self.code_callback(code)



import os
from os.path import join

from json_database import JsonStorageXDG, JsonDatabaseXDG
from ovos_config.config import Configuration
from ovos_utils.configuration import get_xdg_data_save_path


class OAuthTokenDatabase(JsonStorageXDG):
    """ This helper class creates ovos-config-assistant/ovos-backend-manager compatible json databases
        This allows users to use oauth even when not using a backend"""
    def __init__(self):
        super().__init__("ovos_oauth")

    def add_token(self, oauth_service, token_data):
        self[oauth_service] = token_data

    def total_tokens(self):
        return len(self)


class OAuthApplicationDatabase(JsonStorageXDG):
    """ This helper class creates ovos-config-assistant/ovos-backend-manager compatible json databases
        This allows users to use oauth even when not using a backend"""
    def __init__(self):
        super().__init__("ovos_oauth_apps")

    def add_application(self, oauth_service,
                        client_id, client_secret,
                        auth_endpoint, token_endpoint, refresh_endpoint,
                        callback_endpoint, scope, shell_integration=True):
        self[oauth_service] = {"oauth_service": oauth_service,
                               "client_id": client_id,
                               "client_secret": client_secret,
                               "auth_endpoint": auth_endpoint,
                               "token_endpoint": token_endpoint,
                               "refresh_endpoint": refresh_endpoint,
                               "callback_endpoint": callback_endpoint,
                               "scope": scope,
                               "shell_integration": shell_integration}

    def total_apps(self):
        return len(self)


class BackendDatabase:
    """ This helper class creates ovos-config-assistant/ovos-backend-manager compatible json databases
    This allows users to visualize metrics, tag wake words and configure devices
    even when not using a backend"""

    def __init__(self, uuid):
        self.uuid = uuid

    def update_device_db(self, data):
        with JsonStorageXDG("ovos_preferences", subfolder="OpenVoiceOS") as db:
            db.update(data)
        cfg = Configuration()
        tts = cfg.get("tts", {}).get("module")
        ww = cfg.get("listener", {}).get("wake_word", "hey_mycroft")

        with JsonStorageXDG("ovos_devices") as db:
            skips = ["state", "coreVersion", "platform", "platform_build", "enclosureVersion"]
            default = {
                "uuid": self.uuid,
                "isolated_skills": True,
                "name": "LocalDevice",
                "device_location": "127.0.0.1",
                "email": "",
                "date_format": cfg.get("date_format") or "DMY",
                "time_format": cfg.get("time_format") or "full",
                "system_unit": cfg.get("system_unit") or "metric",
                "opt_in": cfg.get("opt_in", False),
                "lang": cfg.get("lang", "en-us"),
                "location": cfg.get("location", {}),
                "default_tts": tts,
                "default_tts_cfg": cfg.get("tts", {}).get(tts, {}),
                "default_ww": ww,
                "default_ww_cfg": cfg.get("hotwords", {}).get(ww, {})
            }
            data = {k: v if k not in data else data[k]
                    for k, v in default.items() if k not in skips}
            db[self.uuid] = data

    def update_metrics_db(self, name, data):
        # shared with personal backend for UI compat
        with JsonDatabaseXDG("ovos_metrics") as db:
            db.add_item({
                "metric_id": len(db) + 1,
                "uuid": self.uuid,
                "metric_type": name,
                "meta": data
            })

    def update_ww_db(self, params):
        listener_config = Configuration().get("listener", {})
        save_path = listener_config.get('save_path', f"{get_xdg_data_save_path()}/listener")
        saved_wake_words_dir = join(save_path, 'wake_words')
        filename = join(saved_wake_words_dir,
                        '_'.join(str(params[k]) for k in sorted(params)) +
                        '.wav')
        if os.path.isfile(filename):
            with JsonDatabaseXDG("ovos_wakewords") as db:
                db.add_item({
                    "wakeword_id": len(db) + 1,
                    "uuid": self.uuid,
                    "meta": params,
                    "path": filename,
                    "transcription": params["name"]
                })
        return filename



from ovos_utils.ovos_service_api import OVOSApiService

from ovos_backend_client.backends.offline import BackendType, AbstractPartialBackend

OVOS_API_URL = "https://api.openvoiceos.com"
OVOS_STT_URL = "https://stt.openvoiceos.com/stt"


class OVOSAPIBackend(AbstractPartialBackend):

    def __init__(self, url=OVOS_API_URL, version="v1", identity_file=None, credentials=None):
        super().__init__(url, version, identity_file, BackendType.OVOS_API, credentials)
        self.client = OVOSApiService()

    @property
    def uuid(self):
        self.check_token()  # register device if needed
        return self.client.get_uuid()

    @property
    def access_token(self):
        self.check_token()
        return self.client.get_session_token()

    @property
    def headers(self):
        return {'session_challenge': self.access_token}

    def check_token(self):
        if not self.client.get_uuid():
            self.client.register_device()

    def refresh_token(self):
        self.check_token()

    # OWM Api
    def owm_get_weather(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        # default to configured location
        lat, lon = lat_lon or self._get_lat_lon()
        reqdata = {"lat": lat,
                   "lon": lon,
                   "units": units,
                   "lang": lang}
        url = f'{self.backend_url}/weather/onecall_weather_report/{self.uuid}'
        response = self.post(url, data=reqdata, headers={"backend": "OWM"})
        if response.status_code != 200:
            raise RuntimeError(f"OWM api failed, status_code {response.status_code}")
        response = response.json()
        if "error" in response:
            raise RuntimeError(response["error"])
        return response

    def owm_get_current(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        # default to configured location
        lat, lon = lat_lon or self._get_lat_lon()
        reqdata = {"lat": lat,
                   "lon": lon,
                   "units": units,
                   "lang": lang}
        url = f'{self.backend_url}/weather/generate_current_weather_report/{self.uuid}'
        response = self.post(url, data=reqdata, headers={"backend": "OWM"})
        if response.status_code != 200:
            raise RuntimeError(f"OWM api failed, status_code {response.status_code}")
        response = response.json()
        if "error" in response:
            raise RuntimeError(response["error"])
        return response

    def owm_get_hourly(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        # default to configured location

        lat, lon = lat_lon or self._get_lat_lon()
        reqdata = {"lat": lat,
                   "lon": lon,
                   "units": units,
                   "lang": lang}
        url = f'{self.backend_url}/weather/generate_hourly_weather_report/{self.uuid}'
        response = self.post(url, data=reqdata, headers={"backend": "OWM"})
        if response.status_code != 200:
            raise RuntimeError(f"OWM api failed, status_code {response.status_code}")
        response = response.json()
        if "error" in response:
            raise RuntimeError(response["error"])
        return response

    def owm_get_daily(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        raise KeyError("daily forecast API is non free")

    # Wolfram Alpha API
    def wolfram_spoken(self, query, units="metric", lat_lon=None, optional_params=None):
        optional_params = optional_params or {}
        if not lat_lon:
            lat_lon = self._get_lat_lon(**optional_params)
        params = {'input': query,
                  "geolocation": "{},{}".format(*lat_lon),
                  'units': units,
                  **optional_params}
        url = f'{self.backend_url}/wolframalpha/spoken/{self.uuid}'
        data = self.post(url=url, params=params)
        return data.text

    def wolfram_simple(self, query, units="metric", lat_lon=None, optional_params=None):
        optional_params = optional_params or {}
        if not lat_lon:
            lat_lon = self._get_lat_lon(**optional_params)
        params = {'input': query,
                  "geolocation": "{},{}".format(*lat_lon),
                  'units': units,
                  **optional_params}
        url = f'{self.backend_url}/wolframalpha/simple/{self.uuid}'
        data = self.get(url=url, params=params)
        return data.text

    def wolfram_full_results(self, query, units="metric", lat_lon=None, optional_params=None):
        """Wrapper for the WolframAlpha Full Results v2 API.
        https://products.wolframalpha.com/api/documentation/
        Pods of interest
        - Input interpretation - Wolfram's determination of what is being asked about.
        - Name - primary name of
        """
        optional_params = optional_params or {}
        if not lat_lon:
            lat_lon = self._get_lat_lon(**optional_params)
        params = {'input': query,
                  "units": units,
                  "mode": "Default",
                  "format": "image,plaintext",
                  "geolocation": "{},{}".format(*lat_lon),
                  "output": "json",
                  **optional_params}
        url = f'{self.backend_url}/wolframalpha/full/{self.uuid}'
        data = self.get(url=url, params=params)
        return data.json()

    # Geolocation Api
    def geolocation_get(self, location):
        """Call the geolocation endpoint.

        Args:
            location (str): the location to lookup (e.g. Kansas City Missouri)

        Returns:
            str: JSON structure with lookup results
        """
        url = f"{self.backend_url}/geolocate/location/config"
        location = self.get(url, params={"address": location})
        location = location.json()
        if "timezone" not in location:
            location["timezone"] = self._get_timezone(
                lon=location["coordinate"]["longitude"],
                lat=location["coordinate"]["latitude"])
        return location

    # STT Api
    def stt_get(self, audio, language="en-us", limit=1):
        """ Web API wrapper for performing Speech to Text (STT)

        Args:
            audio (bytes): The recorded audio, as in a FLAC file
            language (str): A BCP-47 language code, e.g. "en-US"
            limit (int): Maximum alternate transcriptions

       """
        content_type = "audio/x-flac"

        data = self.post(OVOS_STT_URL, data=audio,
                         params={"lang": language, "limit": limit},
                         headers={"Content-Type": content_type})
        if data.status_code == 200:
            return [data.text]
        raise RuntimeError(f"STT api failed, status_code {data.status_code}")

    # Email API
    def email_send(self, title, body, sender):
        # get default recipient from config,
        # in ovos we dont have email tied to user accounts
        mail_config = self.credentials["email"]
        recipient = mail_config.get("recipient") or \
                    mail_config.get("smtp", {}).get("username")

        reqdata = {"recipient": recipient,
                   "subject": subject,
                   "body": body}
        url = f"{self.backend_url}/send/mail/{self.uuid}"
        requests.post(url, data=reqdata, headers=self.headers)



if __name__ == "__main__":
    b = OVOSAPIBackend()
    # a = b.geolocation_get("Fafe")
    a = b.wolfram_full_results("2+2")
    print(a)
    a = b.wolfram_spoken("what is the speed of light")
    print(a)
    a = b.wolfram_simple("what is the speed of light")
    print(a)
    exit()

    from speech_recognition import Recognizer, AudioFile

    with AudioFile("/home/user/PycharmProjects/selene_api/test/test.wav") as source:
        audio = Recognizer().record(source)

    flac_data = audio.get_flac_data()
    # a = b.stt_get(flac_data)

    a = b.owm_get_weather()
    # a = b.owm_get_daily()
    # a = b.owm_get_hourly()
    # a = b.owm_get_current()
    print(a)



from ovos_config.config import Configuration

from ovos_backend_client.backends.personal import PersonalBackend, BackendType

SELENE_API_URL = "https://api.mycroft.ai"
SELENE_PRECISE_URL = "https://training.mycroft.ai/precise/upload"


class SeleneBackend(PersonalBackend):

    def __init__(self, url=SELENE_API_URL, version="v1", identity_file=None, credentials=None):
        super().__init__(url, version, identity_file, credentials)
        self.backend_type = BackendType.SELENE

    # Device API
    def device_upload_wake_word_v1(self, audio, params, upload_url=None):
        """ upload precise wake word V1 endpoint - url can be external to backend"""
        # ensure default value for selene backend
        if not upload_url:
            config = Configuration().get("listener", {}).get("wake_word_upload", {})
            upload_url = config.get("url") or SELENE_PRECISE_URL
        return super().device_upload_wake_word_v1(audio, params, upload_url)

    # Admin API - NOT available, use home.mycroft.ai instead
    def admin_pair(self, uuid=None):
        raise RuntimeError(f"AdminAPI not available for {self.backend_type}")

    def admin_set_device_location(self, uuid, loc):
        raise RuntimeError(f"AdminAPI not available for {self.backend_type}")

    def admin_set_device_prefs(self, uuid, prefs):
        raise RuntimeError(f"AdminAPI not available for {self.backend_type}")

    def admin_set_device_info(self, uuid, info):
        raise RuntimeError(f"AdminAPI not available for {self.backend_type}")



import re
from pprint import pformat

from ovos_utils import camel_case_split
from ovos_utils.log import LOG

from ovos_backend_client.api import DeviceApi
from ovos_backend_client.pairing import is_paired
from ovos_backend_client.identity import IdentityManager


def _is_remote_list(values):
    """Check if list corresponds to a backend formatted collection of dicts
    """
    for v in values:
        if not isinstance(v, dict):
            return False
        if "@type" not in v.keys():
            return False
    return True


def _translate_remote(config, setting):
    """Translate config names from server to equivalents for mycroft-core.

    Args:
        config:     base config to populate
        settings:   remote settings to be translated
    """
    IGNORED_SETTINGS = ["uuid", "@type", "active", "user", "device"]

    for k, v in setting.items():
        if k not in IGNORED_SETTINGS:
            # Translate the CamelCase values stored remotely into the
            # Python-style names used within mycroft-core.
            key = re.sub(r"Setting(s)?", "", k)
            key = camel_case_split(key).replace(" ", "_").lower()
            if isinstance(v, dict):
                config[key] = config.get(key, {})
                _translate_remote(config[key], v)
            elif isinstance(v, list):
                if _is_remote_list(v):
                    if key not in config:
                        config[key] = {}
                    _translate_list(config[key], v)
                else:
                    config[key] = v
            else:
                config[key] = v


def _translate_list(config, values):
    """Translate list formatted by mycroft server.

    Args:
        config (dict): target config
        values (list): list from mycroft server config
    """
    for v in values:
        module = v["@type"]
        if v.get("active"):
            config["module"] = module
        config[module] = config.get(module, {})
        _translate_remote(config[module], v)


class RemoteConfigManager:
    """Config dictionary fetched from mycroft.ai."""

    def __init__(self, url=None, version="v1", identity_file=None):
        self.api = DeviceApi(url, version, identity_file)
        self.config = {"server": {"disabled": True}}

    def download(self):
        if not is_paired(url=self.api.backend_url,
                         version=self.api.backend_version,
                         identity_file=IdentityManager.IDENTITY_FILE):
            self.config = {"server": {"disabled": True}}
            return
        else:
            self.config["server"]["disabled"] = False

        try:
            remote = self.api.get_settings()

            try:  # this call is unnecessary in personal backend but needed in selene
                location = self.api.get_location()
                remote["location"] = location
            except Exception as e:
                LOG.error(f"Exception fetching remote location: {e}")

            # Remove server specific entries
            _translate_remote(self.config, remote)

        except Exception as e:
            LOG.error(f"Exception fetching remote configuration: {e}")

    def print(self):
        print(pformat(self.config))


if __name__ == "__main__":
    cfg = RemoteConfigManager()
    cfg.download()
    cfg.print()
    # {'date_format': 'DMY',
    #  'hotwords': {'hey_mycroft': {'module': 'ovos-ww-plugin-pocketsphinx',
    #                               'phonemes': 'HH EY . M AY K R AO F T',
    #                               'threshold': 1e-90}},
    #  'listener': {'wake_word': 'hey_mycroft'},
    #  'location': {'city': {'code': 'Porto',
    #                        'name': 'Porto',
    #                        'region': {'code': '13',
    #                                   'country': {'code': 'PT', 'name': 'Portugal'},
    #                                   'name': 'Porto'},
    #                        'state': {'code': '13',
    #                                  'country': {'code': 'PT', 'name': 'Portugal'},
    #                                  'name': 'Porto'}},
    #               'coordinate': {'latitude': 41.1691, 'longitude': -8.6793},
    #               'timezone': {'code': 'Europe/Lisbon',
    #                            'dst_offset': 3600000,
    #                            'name': 'Europe/Lisbon',
    #                            'offset': -21600000}},
    #  'opt_in': False,
    #  'server': {'disabled': False},
    #  'system_unit': 'metric',
    #  'time_format': 'full',
    #  'tts': {'module': 'ovos-tts-plugin-mimic2',
    #          'ovos-tts-plugin-mimic2': {'module': 'ovos-tts-plugin-mimic2',
    #                                     'voice': 'kusal'}}}



import json
import time
from io import BytesIO, StringIO
from tempfile import NamedTemporaryFile
from uuid import uuid4

from json_database import JsonStorageXDG
from ovos_config.config import Configuration
from ovos_config.config import update_mycroft_config
from ovos_plugin_manager.stt import OVOSSTTFactory, get_stt_config
from ovos_utils.smtp_utils import send_smtp

from ovos_backend_client.identity import IdentityManager
from ovos_backend_client.backends.base import AbstractBackend, BackendType
from ovos_backend_client.database import BackendDatabase


class OfflineBackend(AbstractBackend):

    def __init__(self, url="127.0.0.1", version="v1", identity_file=None, credentials=None):
        super().__init__(url, version, identity_file, BackendType.OFFLINE, credentials)
        self.stt = None

    # OWM API
    def owm_get_weather(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        # default to configured location

        lat, lon = lat_lon or self._get_lat_lon()
        params = {
            "lang": lang,
            "units": units,
            "lat": lat, "lon": lon,
            "appid": self.credentials["owm"]
        }
        url = "https://api.openweathermap.org/data/2.5/onecall"
        response = self.get(url, params=params)
        return response.json()

    def owm_get_current(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        # default to configured location

        lat, lon = lat_lon or self._get_lat_lon()
        params = {
            "lang": lang,
            "units": units,
            "lat": lat, "lon": lon,
            "appid": self.credentials["owm"]
        }
        url = "https://api.openweathermap.org/data/2.5/weather"
        response = self.get(url, params=params)
        return response.json()

    def owm_get_hourly(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        # default to configured location

        lat, lon = lat_lon or self._get_lat_lon()
        params = {
            "lang": lang,
            "units": units,
            "lat": lat, "lon": lon,
            "appid": self.credentials["owm"]
        }
        url = "https://api.openweathermap.org/data/2.5/forecast"
        response = self.get(url, params=params)
        return response.json()

    def owm_get_daily(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        # default to configured location

        lat, lon = lat_lon or self._get_lat_lon()
        params = {
            "lang": lang,
            "units": units,
            "lat": lat, "lon": lon,
            "appid": self.credentials["owm"]
        }
        url = "https://api.openweathermap.org/data/2.5/forecast/daily"
        response = self.get(url, params=params)
        return response.json()

    # Wolfram Alpha Api
    def wolfram_spoken(self, query, units="metric", lat_lon=None, optional_params=None):
        optional_params = optional_params or {}
        if not lat_lon:
            lat_lon = self._get_lat_lon(**optional_params)
        params = {'i': query,
                  "geolocation": "{},{}".format(*lat_lon),
                  'units': units,
                  **optional_params}
        url = 'https://api.wolframalpha.com/v1/spoken'
        params["appid"] = self.credentials["wolfram"]
        return self.get(url, params=params).text

    def wolfram_simple(self, query, units="metric", lat_lon=None, optional_params=None):
        optional_params = optional_params or {}
        if not lat_lon:
            lat_lon = self._get_lat_lon(**optional_params)
        params = {'i': query,
                  "geolocation": "{},{}".format(*lat_lon),
                  'units': units,
                  **optional_params}
        url = 'https://api.wolframalpha.com/v1/simple'
        params["appid"] = self.credentials["wolfram"]
        return self.get(url, params=params).text

    def wolfram_full_results(self, query, units="metric", lat_lon=None, optional_params=None):
        """Wrapper for the WolframAlpha Full Results v2 API.
        https://products.wolframalpha.com/api/documentation/
        Pods of interest
        - Input interpretation - Wolfram's determination of what is being asked about.
        - Name - primary name of
        """
        optional_params = optional_params or {}
        if not lat_lon:
            lat_lon = self._get_lat_lon(**optional_params)
        params = {'input': query,
                  "units": units,
                  "mode": "Default",
                  "format": "image,plaintext",
                  "geolocation": "{},{}".format(*lat_lon),
                  "output": "json",
                  **optional_params}
        url = 'https://api.wolframalpha.com/v2/query'
        params["appid"] = self.credentials["wolfram"]
        data = self.get(url, params=params)
        return data.json()

    # Geolocation Api
    def geolocation_get(self, location):
        """Call the geolocation endpoint.

        Args:
            location (str): the location to lookup (e.g. Kansas City Missouri)

        Returns:
            str: JSON structure with lookup results
        """
        url = "https://nominatim.openstreetmap.org/search"
        data = self.get(url, params={"q": location, "format": "json", "limit": 1}).json()[0]
        url = "https://nominatim.openstreetmap.org/details.php?osmtype=W&osmid=38210407&format=json"
        details = self.get(url, params={"osmid": data['osm_id'], "osmtype": data['osm_type'][0].upper(),
                                        "format": "json"}).json()

        location = {
            "city": {
                "code": details["addresstags"].get("postcode") or details["calculated_postcode"] or "",
                "name": details["localname"],
                "state": {
                    "code": details["addresstags"].get("state_code") or details["calculated_postcode"] or "",
                    "name": details["addresstags"].get("state") or data["display_name"].split(", ")[0],
                    "country": {
                        "code": details["country_code"].upper() or details["addresstags"].get("country"),
                        "name": data["display_name"].split(", ")[-1]
                    }
                }
            },
            "coordinate": {
                "latitude": data["lat"],
                "longitude": data["lon"]
            }
        }
        if "timezone" not in location:
            location["timezone"] = self._get_timezone(
                lon=location["coordinate"]["longitude"],
                lat=location["coordinate"]["latitude"])
        return location

    # Device Api
    def device_get(self):
        """ Retrieve all device information from the web backend """
        data = JsonStorageXDG("ovos_device_info.json", subfolder="OpenVoiceOS")
        for k, v in super().device_get().items():
            if k not in data:
                data[k] = v
        return data

    def device_get_settings(self):
        """ Retrieve device settings information from the web backend

        Returns:
            str: JSON string with user configuration information.
        """
        return Configuration()  # TODO format keys or not needed ?

    def device_get_skill_settings_v1(self):
        """ old style bidirectional skill settings api, still available!"""
        # TODO scan skill xdg paths
        return []

    def device_put_skill_settings_v1(self, data=None):
        """ old style bidirectional skill settings api, still available!"""
        # do nothing, skills manage their own settings lifecycle
        return {}

    def device_get_code(self, state=None):
        return "ABCDEF"  # dummy data

    def device_activate(self, state, token,
                        core_version="unknown",
                        platform="unknown",
                        platform_build="unknown",
                        enclosure_version="unknown"):
        data = {"state": state,
                "token": token,
                "coreVersion": core_version,
                "platform": platform,
                "platform_build": platform_build,
                "enclosureVersion": enclosure_version}
        identity = self.admin_pair(state)
        data["uuid"] = data.pop("state")
        data["token"] = self.access_token
        BackendDatabase(self.uuid).update_device_db(data)
        db = JsonStorageXDG("ovos_device_info.json", subfolder="OpenVoiceOS")
        db.update(data)
        db.store()
        return identity

    def device_update_version(self,
                              core_version="unknown",
                              platform="unknown",
                              platform_build="unknown",
                              enclosure_version="unknown"):
        data = {"coreVersion": core_version,
                "platform": platform,
                "platform_build": platform_build,
                "enclosureVersion": enclosure_version,
                "token": self.access_token}
        BackendDatabase(self.uuid).update_device_db(data)
        db = JsonStorageXDG("ovos_device_info.json", subfolder="OpenVoiceOS")
        db.update(data)
        db.store()

    def device_report_metric(self, name, data):
        return self.metrics_upload(name, data)

    def device_get_location(self):
        """ Retrieve device location information from the web backend

        Returns:
            str: JSON string with user location.
        """
        return Configuration().get("location") or {}

    def device_get_oauth_token(self, dev_cred):
        """
            Get Oauth token for dev_credential dev_cred.

            Argument:
                dev_cred:   development credentials identifier

            Returns:
                json string containing token and additional information
        """
        raise self.oauth_get_token(dev_cred)

    def device_get_skill_settings(self):
        """Get the remote skill settings for all skills on this device."""
        # TODO - scan xdg paths ?
        return {}

    def device_upload_skill_metadata(self, settings_meta):
        """Upload skill metadata.

        Args:
            settings_meta (dict): skill info and settings in JSON format
        """
        # Do nothing, skills manage their own settingsmeta.json files
        return

    def device_upload_skills_data(self, data):
        """ Upload skills.json file. This file contains a manifest of installed
        and failed installations for use with the Marketplace.

        Args:
             data: dictionary with skills data from msm
        """
        with JsonStorageXDG("ovos_skills_meta.json", subfolder="OpenVoiceOS") as db:
            db.update(data)

    def device_upload_wake_word_v1(self, audio, params, upload_url=None):
        """ upload precise wake word V1 endpoint - url can be external to backend"""
        return self.dataset_upload_wake_word(audio, params, upload_url)

    def device_upload_wake_word(self, audio, params):
        """ upload precise wake word V2 endpoint - integrated with device api"""
        return self.dataset_upload_wake_word(audio, params)

    # Metrics API
    def metrics_upload(self, name, data):
        """ upload metrics"""
        BackendDatabase(self.uuid).update_metrics_db(name, data)
        return {}

    # Dataset API
    def dataset_upload_wake_word(self, audio, params, upload_url=None):
        """ upload wake word sample - url can be external to backend"""
        if Configuration().get("listener", {}).get('record_wake_words'):
            BackendDatabase(self.uuid).update_ww_db(params)  # update metadata db for ww tagging UI

        upload_url = upload_url or Configuration().get("listener", {}).get("wake_word_upload", {}).get("url")
        if upload_url:
            # upload to arbitrary server
            ww_files = {
                'audio': BytesIO(audio.get_wav_data()),
                'metadata': StringIO(json.dumps(params))
            }
            return self.post(upload_url, files=ww_files)
        return {}

    # Email API
    def email_send(self, title, body, sender):
        """ will raise KeyError if SMTP not configured in mycroft.conf"""
        body += f"\n\nsent by: {sender}"  # append skill_id info to body

        mail_config = self.credentials["email"]

        smtp_config = mail_config["smtp"]
        user = smtp_config["username"]
        pswd = smtp_config["password"]
        host = smtp_config["host"]
        port = smtp_config.get("port", 465)

        recipient = mail_config.get("recipient") or user

        send_smtp(user, pswd,
                  user, recipient,
                  title, body,
                  host, port)

    # OAuth API
    def oauth_get_token(self, dev_cred):
        """
            Get Oauth token for dev_credential dev_cred.

            Argument:
                dev_cred:   development credentials identifier

            Returns:
                json string containing token and additional information
        """
        return JsonStorageXDG("ovos_oauth").get(dev_cred) or {}

    # Admin API
    def admin_pair(self, uuid=None):
        uuid = uuid or str(uuid4())
        # create dummy identity file for third parties expecting it for pairing checks
        identity = {"uuid": uuid,
                    "access": "OVOSdbF1wJ4jA5lN6x6qmVk_QvJPqBQZTUJQm7fYzkDyY_Y=",
                    "refresh": "OVOS66c5SpAiSpXbpHlq9HNGl1vsw_srX49t5tCv88JkhuE=",
                    "expires_at": time.time() + 9999999999}
        # save identity file
        IdentityManager.save(identity)
        return identity

    def admin_set_device_location(self, uuid, loc):
        """
        loc = {
            "city": {
                "code": "Lawrence",
                "name": "Lawrence",
                "state": {
                    "code": "KS",
                    "name": "Kansas",
                    "country": {
                        "code": "US",
                        "name": "United States"
                    }
                }
            },
            "coordinate": {
                "latitude": 38.971669,
                "longitude": -95.23525
            },
            "timezone": {
                "code": "America/Chicago",
                "name": "Central Standard Time",
                "dstOffset": 3600000,
                "offset": -21600000
            }
        }
        """
        update_mycroft_config({"location": loc})

    def admin_set_device_prefs(self, uuid, prefs):
        """
        prefs = {"time_format": "full",
                "date_format": "DMY",
                "system_unit": "metric",
                "lang": "en-us",
                "wake_word": "hey_mycroft",
                "ww_config": {"phonemes": "HH EY . M AY K R AO F T",
                             "module": "ovos-ww-plugin-pocketsphinx",
                             "threshold": 1e-90},
                "tts_module": "ovos-tts-plugin-mimic",
                "tts_config": {"voice": "ap"}}
        """
        with JsonStorageXDG("ovos_device_info.json", subfolder="OpenVoiceOS") as db:
            db.update(prefs)
        cfg = dict(prefs)
        cfg["listener"] = {}
        cfg["hotwords"] = {}
        cfg["tts"] = {}
        tts = None
        tts_cfg = {}
        ww = None
        ww_cfg = {}
        if "wake_word" in cfg:
            ww = cfg.pop("wake_word")
        if "ww_config" in cfg:
            ww_cfg = cfg.pop("ww_config")
        if "tts_module" in cfg:
            tts = cfg.pop("tts_module")
        if "tts_config" in cfg:
            tts_cfg = cfg.pop("tts_config")
            if not tts:
                tts = tts_cfg.get("module")
        if tts:
            cfg["tts"]["module"] = tts
            cfg["tts"][tts] = tts_cfg
        if ww:
            cfg["listener"]["wake_word"] = ww
            cfg["hotwords"][ww] = ww_cfg
        update_mycroft_config(cfg)

    def admin_set_device_info(self, uuid, info):
        """
        info = {"opt_in": True,
                "name": "my_device",
                "device_location": "kitchen",
                "email": "notifications@me.com",
                "isolated_skills": False,
                "lang": "en-us"}
        """
        update_mycroft_config({"opt_in": info["opt_in"], "lang": info["lang"]})
        with JsonStorageXDG("ovos_device_info.json", subfolder="OpenVoiceOS") as db:
            db.update(info)

    # STT Api
    def load_stt_plugin(self, config=None, lang=None):
        config = config or get_stt_config(config)
        if lang:
            config["lang"] = lang
        self.stt = OVOSSTTFactory.create(config)

    def stt_get(self, audio, language="en-us", limit=1):
        """ Web API wrapper for performing Speech to Text (STT)

        Args:
            audio (bytes): The recorded audio, as in a FLAC file
            language (str): A BCP-47 language code, e.g. "en-US"
            limit (int): Maximum alternate transcriptions

       """
        if self.stt is None:
            self.load_stt_plugin(lang=language)
        with NamedTemporaryFile() as fp:
            fp.write(audio)
            with AudioFile(fp.name) as source:
                audio = Recognizer().record(source)
        tx = self.stt.execute(audio, language)
        if isinstance(tx, str):
            tx = [tx]
        return tx


class AbstractPartialBackend(OfflineBackend):
    """ helper class that internally delegates unimplemented methods to offline backend implementation
    backends that only provide microservices and no DeviceApi should subclass from here
    """

    def __init__(self, url=None, version="v1", identity_file=None, backend_type=BackendType.OFFLINE, credentials=None):
        super().__init__(url, version, identity_file, credentials)
        self.backend_type = backend_type


if __name__ == "__main__":
    b = OfflineBackend()
    b.load_stt_plugin({"module": "ovos-stt-plugin-vosk"})
    # a = b.geolocation_get("Fafe")
    # a = b.wolfram_full_results("2+2")
    # a = b.wolfram_spoken("what is the speed of light")
    # a = b.owm_get_weather()

    from speech_recognition import Recognizer, AudioFile

    with AudioFile("/home/user/PycharmProjects/selene_api/test/test.wav") as source:
        audio = Recognizer().record(source)

    flac_data = audio.get_flac_data()
    a = b.stt_get(flac_data)

    # a = b.owm_get_weather()
    # a = b.owm_get_daily()
    # a = b.owm_get_hourly()
    # a = b.owm_get_current()
    print(a)



import json
import os
import time
from io import BytesIO, StringIO

from ovos_config.config import Configuration
from ovos_utils.log import LOG
from requests.exceptions import HTTPError

from ovos_backend_client.backends.offline import AbstractPartialBackend, BackendType
from ovos_backend_client.identity import IdentityManager, identity_lock
import requests


class PersonalBackend(AbstractPartialBackend):

    def __init__(self, url="http://0.0.0.0:6712", version="v1", identity_file=None, credentials=None):
        super().__init__(url, version, identity_file, BackendType.PERSONAL, credentials)

    def refresh_token(self):
        try:
            identity_lock.acquire(blocking=False)
            # NOTE: requests needs to be used instead of self.get due to self.get calling this
            data = requests.get(f"{self.backend_url}/{self.backend_version}/auth/token", headers=self.headers).json()
            IdentityManager.save(data, lock=False)
            LOG.debug('Saved credentials')
        except:
            LOG.warning("Failed to refresh access token")
        finally:
            try:
                identity_lock.release()
            except RuntimeError:  # RuntimeError: release unlocked lock
                pass

    # OWM Api
    def owm_get_weather(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        # default to configured location
        lat, lon = lat_lon or self._get_lat_lon()
        response = self.get(url=f"{self.backend_url}/{self.backend_version}/owm/onecall",
                            params={
                                "lang": self.owm_language(lang),
                                "lat": lat,
                                "lon": lon,
                                "units": units})
        return response.json()

    def owm_get_current(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        # default to configured location
        lat, lon = lat_lon or self._get_lat_lon()
        response = self.get(url=f"{self.backend_url}/{self.backend_version}/owm/weather",
                            params={
                                "lang": self.owm_language(lang),
                                "lat": lat,
                                "lon": lon,
                                "units": units})
        return response.json()

    def owm_get_hourly(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        # default to configured location
        lat, lon = lat_lon or self._get_lat_lon()
        response = self.get(url=f"{self.backend_url}/{self.backend_version}/owm/forecast",
                            params={
                                "lang": self.owm_language(lang),
                                "lat": lat,
                                "lon": lon,
                                "units": units})
        return response.json()

    def owm_get_daily(self, lat_lon=None, lang="en-us", units="metric"):
        """Issue an API call and map the return value into a weather report

        Args:
            units (str): metric or imperial measurement units
            lat_lon (tuple): the geologic (latitude, longitude) of the weather location
        """
        # default to configured location
        lat, lon = lat_lon or self._get_lat_lon()
        response = self.get(url=f"{self.backend_url}/{self.backend_version}/owm/forecast/daily",
                            params={
                                "lang": self.owm_language(lang),
                                "lat": lat,
                                "lon": lon,
                                "units": units})
        return response.json()

    # Wolfram Alpha API
    def wolfram_spoken(self, query, units="metric", lat_lon=None, optional_params=None):
        optional_params = optional_params or {}
        if not lat_lon:
            lat_lon = self._get_lat_lon(**optional_params)
        params = {'i': query,
                  "geolocation": "{},{}".format(*lat_lon),
                  'units': units,
                  **optional_params}
        url = f"{self.backend_url}/{self.backend_version}/wolframAlphaSpoken"
        data = self.get(url=url, params=params)
        return data.text

    def wolfram_simple(self, query, units="metric", lat_lon=None, optional_params=None):
        optional_params = optional_params or {}
        if not lat_lon:
            lat_lon = self._get_lat_lon(**optional_params)
        params = {'i': query,
                  "geolocation": "{},{}".format(*lat_lon),
                  'units': units,
                  **optional_params}
        url = f"{self.backend_url}/{self.backend_version}/wolframAlphaSimple"
        data = self.get(url=url, params=params)
        return data.text

    def wolfram_full_results(self, query, units="metric", lat_lon=None, optional_params=None):
        """Wrapper for the WolframAlpha Full Results v2 API.
        https://products.wolframalpha.com/api/documentation/
        Pods of interest
        - Input interpretation - Wolfram's determination of what is being asked about.
        - Name - primary name of
        """
        optional_params = optional_params or {}
        if not lat_lon:
            lat_lon = self._get_lat_lon(**optional_params)
        params = {'input': query,
                  "units": units,
                  "mode": "Default",
                  "format": "image,plaintext",
                  "geolocation": "{},{}".format(*lat_lon),
                  "output": "json",
                  **optional_params}
        url = f"{self.backend_url}/{self.backend_version}/wolframAlphaFull"
        data = self.get(url=url, params=params)
        return data.json()

    # Geolocation Api
    def geolocation_get(self, location):
        """Call the geolocation endpoint.

        Args:
            location (str): the location to lookup (e.g. Kansas City Missouri)

        Returns:
            str: JSON structure with lookup results
        """
        url = f"{self.backend_url}/{self.backend_version}/geolocation"
        location = self.get(url, params={"location": location}).json()['data']
        if "timezone" not in location:
            location["timezone"] = self._get_timezone(
                lon=location["coordinate"]["longitude"],
                lat=location["coordinate"]["latitude"])
        return location

    # STT Api
    def stt_get(self, audio, language="en-us", limit=1):
        """ Web API wrapper for performing Speech to Text (STT)

        Args:
            audio (bytes): The recorded audio, as in a FLAC file
            language (str): A BCP-47 language code, e.g. "en-US"
            limit (int): Maximum alternate transcriptions

        Returns:
            dict: JSON structure with transcription results
        """
        data = self.post(url=f"{self.backend_url}/{self.backend_version}/stt",
                         data=audio, params={"lang": language, "limit": limit},
                         headers={"Content-Type": "audio/x-flac"})
        if data.status_code == 200:
            return data.json()
        raise RuntimeError(f"STT api failed, status_code {data.status_code}")

    # Device Api
    def device_get(self):
        """ Retrieve all device information from the web backend """
        return self.get(f"{self.backend_url}/{self.backend_version}/device/{self.uuid}").json()

    def device_get_settings(self):
        """ Retrieve device settings information from the web backend

        Returns:
            str: JSON string with user configuration information.
        """
        return self.get(f"{self.backend_url}/{self.backend_version}/device/{self.uuid}/setting").json()

    def device_get_skill_settings_v1(self):
        """ old style bidirectional skill settings api, still available!"""
        return self.get(f"{self.backend_url}/{self.backend_version}/device/{self.uuid}/skill").json()

    def device_put_skill_settings_v1(self, data=None):
        """ old style bidirectional skill settings api, still available!"""
        self.put(f"{self.backend_url}/{self.backend_version}/device/{self.uuid}/skill", json=data)

    def device_get_code(self, state=None):
        state = state or self.uuid
        return self.get(f"{self.backend_url}/{self.backend_version}/device/code", params={"state": state}).json()

    def device_activate(self, state, token,
                        core_version="unknown",
                        platform="unknown",
                        platform_build="unknown",
                        enclosure_version="unknown"):
        data = {"state": state,
                "token": token,
                "coreVersion": core_version,
                "platform": platform,
                "platform_build": platform_build,
                "enclosureVersion": enclosure_version}
        r = self.post(f"{self.backend_url}/{self.backend_version}/device/activate", json=data)
        try:
            return r.json()
        except:
            # raise expected exception handled by pairing manager, any other resets pairing process
            raise HTTPError(f"Device activation failed! {r.status_code}")

    def device_update_version(self,
                              core_version="unknown",
                              platform="unknown",
                              platform_build="unknown",
                              enclosure_version="unknown"):
        data = {"coreVersion": core_version,
                "platform": platform,
                "platform_build": platform_build,
                "enclosureVersion": enclosure_version}
        return self.patch(f"{self.backend_url}/{self.backend_version}/device/{self.uuid}", json=data)

    def device_report_metric(self, name, data):
        return self.post(f"{self.backend_url}/{self.backend_version}/device/{self.uuid}/metric/" + name,
                         json=data)

    def device_get_location(self):
        """ Retrieve device location information from the web backend

        Returns:
            str: JSON string with user location.
        """
        return self.get(f"{self.backend_url}/{self.backend_version}/device/{self.uuid}/location").json()

    def device_get_subscription(self):
        """
            Get information about type of subscription this unit is connected
            to.

            Returns: dictionary with subscription information
        """
        return self.get(f"{self.backend_url}/{self.backend_version}/device/{self.uuid}/subscription").json()

    def device_get_subscriber_voice_url(self, voice=None, arch=None):
        archs = {'x86_64': 'x86_64', 'armv7l': 'arm', 'aarch64': 'arm'}
        arch = arch or archs.get(os.uname()[4])
        if arch:
            url = f"{self.backend_url}/{self.backend_version}/device/{self.uuid}/voice"
            return self.get(url, params={"arch": arch}).json().get('link')

    def device_get_oauth_token(self, dev_cred):
        """
            Get Oauth token for dev_credential dev_cred.

            Argument:
                dev_cred:   development credentials identifier

            Returns:
                json string containing token and additional information
        """
        return self.oauth_get_token(dev_cred)

    def device_get_skill_settings(self):
        """Get the remote skill settings for all skills on this device."""
        return self.get(f"{self.backend_url}/{self.backend_version}/device/{self.uuid}/skill/settings").json()

    def device_send_email(self, title, body, sender):
        return self.put(f"{self.backend_url}/{self.backend_version}/device/{self.uuid}/message",
                        json={"title": title, "body": body, "sender": sender}).json()

    def device_upload_skill_metadata(self, settings_meta):
        """Upload skill metadata.

        Args:
            settings_meta (dict): skill info and settings in JSON format
        """
        return self.put(url=f"{self.backend_url}/{self.backend_version}/device/{self.uuid}/settingsMeta",
                        json=settings_meta)

    def device_upload_skills_data(self, data):
        """ Upload skills.json file. This file contains a manifest of installed
        and failed installations for use with the Marketplace.

        Args:
             data: dictionary with skills data from msm
        """
        return self.put(url=f"{self.backend_url}/{self.backend_version}/device/{self.uuid}/skillJson",
                        json=data)

    def device_upload_wake_word_v1(self, audio, params, upload_url=None):
        """ upload precise wake word V1 endpoint - url can be external to backend"""
        if not upload_url:
            config = Configuration().get("listener", {}).get("wake_word_upload", {})
            upload_url = config.get("url") or f"{self.backend_url}/precise/upload"
        ww_files = {
            'audio': BytesIO(audio.get_wav_data()),
            'metadata': StringIO(json.dumps(params))
        }
        return self.post(upload_url, files=ww_files)

    def device_upload_wake_word(self, audio, params):
        """ upload precise wake word V2 endpoint - integrated with device api"""
        url = f"{self.backend_url}/{self.backend_version}/device/{self.uuid}/wake-word-file"
        request_data = dict(
            wake_word=params['name'],
            engine=params.get('engine_name') or params.get('engine'),
            timestamp=params.get('timestamp') or params.get('time') or str(int(1000 * time.time())),
            model=params['model']
        )
        ww_files = {
            'audio': BytesIO(audio.get_wav_data()),
            'metadata': StringIO(json.dumps(request_data))
        }
        return self.post(url, files=ww_files)

    # Metrics API
    def metrics_upload(self, name, data):
        """ upload metrics"""
        return self.device_report_metric(name, data)

    # Dataset API
    def dataset_upload_wake_word(self, audio, params, upload_url=None):
        """ upload wake word sample - url can be external to backend"""
        if upload_url:  # explicit upload endpoint requested
            return self.device_upload_wake_word_v1(audio, params, upload_url)
        return self.device_upload_wake_word(audio, params)

    # OAuth API
    def oauth_get_token(self, dev_cred):
        """
            Get Oauth token for dev_credential dev_cred.

            Argument:
                dev_cred:   development credentials identifier

            Returns:
                json string containing token and additional information
        """
        return self.get(f"{self.backend_url}/{self.backend_version}/device/{self.uuid}/token/{dev_cred}").json()

    # Email API
    def email_send(self, title, body, sender):
        return self.device_send_email(title, body, sender)

    # Admin API
    def admin_pair(self, uuid=None):
        identity = self.get(f"{self.backend_url}/{self.backend_version}/admin/{uuid}/pair",
                            headers={"Authorization": f"Bearer {self.credentials['admin']}"})
        # save identity file
        self.identity.save(identity)
        return identity

    def admin_set_device_location(self, uuid, loc):
        """
        loc = {
            "city": {
                "code": "Lawrence",
                "name": "Lawrence",
                "state": {
                    "code": "KS",
                    "name": "Kansas",
                    "country": {
                        "code": "US",
                        "name": "United States"
                    }
                }
            },
            "coordinate": {
                "latitude": 38.971669,
                "longitude": -95.23525
            },
            "timezone": {
                "code": "America/Chicago",
                "name": "Central Standard Time",
                "dstOffset": 3600000,
                "offset": -21600000
            }
        }
        """
        return self.put(f"{self.backend_url}/{self.backend_version}/admin/{uuid}/location", json=loc,
                        headers={"Authorization": f"Bearer {self.credentials['admin']}"})

    def admin_set_device_prefs(self, uuid, prefs):
        """
        prefs = {"time_format": "full",
                "date_format": "DMY",
                "system_unit": "metric",
                "lang": "en-us",
                "wake_word": "hey_mycroft",
                "ww_config": {"phonemes": "HH EY . M AY K R AO F T",
                             "module": "ovos-ww-plugin-pocketsphinx",
                             "threshold": 1e-90},
                "tts_module": "ovos-tts-plugin-mimic",
                "tts_config": {"voice": "ap"}}
        """
        return self.put(f"{self.backend_url}/{self.backend_version}/admin/{uuid}/prefs", json=prefs,
                        headers={"Authorization": f"Bearer {self.credentials['admin']}"})

    def admin_set_device_info(self, uuid, info):
        """
        info = {"opt_in": True,
                "name": "my_device",
                "device_location": "kitchen",
                "email": "notifications@me.com",
                "isolated_skills": False,
                "lang": "en-us"}
        """
        return self.put(f"{self.backend_url}/{self.backend_version}/admin/{uuid}/device", json=info,
                        headers={"Authorization": f"Bearer {self.credentials['admin']}"})



import json
import os
import shutil
import time
from os.path import isfile, dirname, expanduser

from combo_lock import ComboLock
from ovos_utils.configuration import get_xdg_config_save_path, get_xdg_base
from ovos_utils.log import LOG

identity_lock = ComboLock('/tmp/identity-lock')


def find_identity():
    locations = [
        IdentityManager.OLD_IDENTITY_FILE,  # old location
        IdentityManager.IDENTITY_FILE,  # xdg location
        "~/mycroft-config/identity/identity2.json",  # smartgic docker default loc
    ]
    for loc in locations:
        loc = expanduser(loc)
        if isfile(loc):
            return loc
    return None


def load_identity():
    locations = [
        IdentityManager.OLD_IDENTITY_FILE,  # old location
        IdentityManager.IDENTITY_FILE,  # xdg location
        "~/mycroft-config/identity/identity2.json",  # smartgic docker default loc
    ]
    for loc in locations:
        loc = expanduser(loc)
        if isfile(loc):
            LOG.debug(f"identity found: {loc}")
            try:
                with open(loc) as f:
                    return json.load(f)
            except:
                LOG.error("invalid identity file!")
                continue
    return {}


class DeviceIdentity:
    def __init__(self, **kwargs):
        self.uuid = kwargs.get("uuid", "")
        self.access = kwargs.get("access", "")
        self.refresh = kwargs.get("refresh", "")
        self.expires_at = kwargs.get("expires_at", -1)

    def is_expired(self):
        return self.refresh and 0 < self.expires_at <= time.time()

    def has_refresh(self):
        return self.refresh != ""


class IdentityManager:
    IDENTITY_FILE = f"{get_xdg_config_save_path()}/identity/identity2.json"
    OLD_IDENTITY_FILE = expanduser(f"~/.{get_xdg_base()}/identity/identity2.json")
    __identity = None

    @classmethod
    def set_identity_file(cls, identity_path):
        cls.IDENTITY_FILE = identity_path
        cls.load()

    @staticmethod
    def _load():
        if isfile(IdentityManager.OLD_IDENTITY_FILE) and \
                not isfile(IdentityManager.IDENTITY_FILE):
            os.makedirs(dirname(IdentityManager.IDENTITY_FILE), exist_ok=True)
            shutil.move(IdentityManager.OLD_IDENTITY_FILE, IdentityManager.IDENTITY_FILE)
        if isfile(IdentityManager.IDENTITY_FILE):
            LOG.debug(f'Loading identity: {IdentityManager.IDENTITY_FILE}')
            try:
                with open(IdentityManager.IDENTITY_FILE) as f:
                    IdentityManager.__identity = DeviceIdentity(**json.load(f))
                return
            except Exception:
                pass
        IdentityManager.__identity = DeviceIdentity()

    @staticmethod
    def load(lock=True):
        try:
            if lock:
                identity_lock.acquire()
                IdentityManager._load()
        finally:
            if lock:
                identity_lock.release()
        return IdentityManager.__identity

    @staticmethod
    def save(login=None, lock=True):
        LOG.debug('Saving identity')
        if lock:
            identity_lock.acquire()
        try:
            if login:
                IdentityManager._update(login)

            os.makedirs(dirname(IdentityManager.IDENTITY_FILE), exist_ok=True)

            with open(IdentityManager.IDENTITY_FILE, "w") as f:
                json.dump(IdentityManager.__identity.__dict__, f)
                f.flush()
                os.fsync(f.fileno())
        finally:
            if lock:
                identity_lock.release()

    @staticmethod
    def _update(login=None):
        LOG.debug('Updating identity')
        login = login or {}
        expiration = login.get("expiration", -1)
        IdentityManager.__identity.uuid = login.get("uuid", "")
        IdentityManager.__identity.access = login.get("accessToken", "")
        IdentityManager.__identity.refresh = login.get("refreshToken", "")
        if expiration > 0:
            IdentityManager.__identity.expires_at = time.time() + expiration
        else:
            IdentityManager.__identity.expires_at = -1

    @staticmethod
    def update(login=None, lock=True):
        if lock:
            identity_lock.acquire()
        try:
            IdentityManager._update(login)
        finally:
            if lock:
                identity_lock.release()

    @staticmethod
    def get():
        if not IdentityManager.__identity:
            IdentityManager.load()
        return IdentityManager.__identity
